import logging

from edos.cli.do.cluster import cluster
from edos.cli.do.database import database
from edos.cli.do.spaces import spaces
from edos.cli.docker import docker
from edos.cli.invoice import invoice

LOG = logging.getLogger(__name__)


def register_commands(group):
    group.add_command(cluster)
    group.add_command(database)
    group.add_command(spaces)
    group.add_command(docker)
    group.add_command(invoice)



import dataclasses
from typing import Optional

from edos.api.digitalocean.databases.models.settings import MysqlSettings


@dataclasses.dataclass
class DatabaseUser:
    name: str
    role: str
    password: str
    mysql_settings: Optional[MysqlSettings] = None



from time import sleep
from typing import Union

import docker
import docker.errors
from docker.models.secrets import Secret
from docker.models.services import Service
from docker.types import SecretReference

from edos.exceptions import (
    ContainerDoesNotExists,
    SecretCreationError,
    SecretDoesNotExists,
    SecretRemovalError,
    ServiceDoesNotExists,
    TaskDoesNotExists,
)
from edos.settings import conf


class DockerService:
    def __init__(self):
        self.client = docker.DockerClient(base_url=conf.DOCKER_SSH_BASE_URL, use_ssh_client=True)

    def get_node_name_for_service(self, service_name: str) -> Union[str, None]:
        service = self.get_service_by_service_name(service_name)
        task = self.get_service_running_task(service)
        node_id = task.get("NodeID")
        return self.client.nodes.get(node_id).attrs.get("Description").get("Hostname")

    def get_service_running_task(self, service: Service) -> dict:
        tasks = service.tasks(filters={"desired-state": "running"})
        if not tasks:
            raise TaskDoesNotExists("No running tasks for this service")
        return tasks[0]

    def get_service_by_service_name(self, service_name: str) -> Service:
        services = self.client.services.list(filters={"name": service_name})
        # service could have similar names, so we need to check if the name is same
        if not services or services[0].name != service_name:
            raise ServiceDoesNotExists("No service found")
        return services[0]

    def get_container(self, hostname: str, service_name: str) -> str:
        client = docker.DockerClient(base_url=f"ssh://{hostname}", use_ssh_client=True)
        service = self.get_service_by_service_name(service_name)
        return client.containers.list(filters={"label": [f"com.docker.swarm.service.id={service.id}"]})[0].name

    def get_services(self) -> list[Service]:
        return self.client.services.list()

    def get_secrets(self) -> list[Secret]:
        return self.client.secrets.list()

    def create_secret(self, name: str, secret_data: bytes) -> Secret:
        """Create a secret, returning Secret object"""
        try:
            return self.client.secrets.create(name=name, data=secret_data)
        except docker.errors.APIError as exc:
            raise SecretCreationError(f"Error when creating secret: {exc.response.reason}")

    def remove_secret(self, name: str):
        """Remove a secret"""
        try:
            self.client.secrets.get(name).remove()
        except docker.errors.NotFound:
            raise SecretDoesNotExists("Secret not found")
        except docker.errors.APIError as exc:
            raise SecretRemovalError(f"Error when removing secret: {exc.response.reason}")

    def get_secret_by_name(self, secret_name: str):
        secrets = self.client.secrets.list(filters={"name": secret_name})
        if not secrets or secrets[0].name != secret_name:
            raise SecretDoesNotExists("No such secret")
        return secrets[0]

    def create_service_with_secret(self, secret: Secret):
        secret_ref = SecretReference(secret_name=secret.name, secret_id=secret.id)
        return self.client.services.create(
            image="alpine",
            secrets=[
                secret_ref,
            ],
            command="tail -f /dev/null",
            constraints=[f"node.hostname=={conf.DOCKER_MAIN_SWARM_HOSTNAME}"],
        )

    def get_container_name_by_service(self, service: Service) -> str:
        container_name = None
        i = 0
        # container could not be created, so we are trying every 10 seconds to get it
        while i != 10:
            sleep(1)
            try:
                container_name = self.client.containers.list(
                    filters={"label": [f"com.docker.swarm.service.id={service.id}"]}
                )[0].name
                return container_name
            except IndexError:
                i = i + 1
        if not container_name:
            raise ContainerDoesNotExists("Timeout")

    def get_services_without_reservation(self) -> list[Service]:
        services = self.client.services.list()
        services_without_reservation: list[Service] = []
        for service in services:
            try:
                service.attrs["Spec"]["TaskTemplate"]["Resources"]["Reservations"]["MemoryBytes"]
            except KeyError:
                services_without_reservation.append(service)
        return services_without_reservation

    def get_sum_of_reservation(self) -> int:
        services = self.client.services.list()
        sum_of_reservation: int = 0
        for service in services:
            if "_cron" in service.name or "_redis" in service.name:
                continue

            try:
                if "node.role == manager" in service.attrs["Spec"]["TaskTemplate"]["Placement"]["Constraints"]:
                    continue
            except KeyError:
                pass

            try:
                sum_of_reservation += service.attrs["Spec"]["TaskTemplate"]["Resources"]["Reservations"]["MemoryBytes"]
            except KeyError:
                pass
        return sum_of_reservation



from edos.api.swarmpit.base_api import SwarmpitAPI
from edos.api.swarmpit.models.memory_response import MemoryResponse


class ServiceSwarmpitAPI(SwarmpitAPI):
    path = "/api/tasks"

    def get_memory_timeseries(self) -> list[MemoryResponse]:
        url = self.build_url()
        resp = self.get_request(url)
        memory_timeseries = []
        for item in resp:
            if item["desiredState"] == "running" and "redis" not in item["serviceName"]:
                memory_timeseries.append(
                    MemoryResponse(
                        item["serviceName"],
                        item["resources"]["reservation"]["memory"],
                        item["stats"]["memory"],
                    )
                )
        return memory_timeseries



import logging

import click

from edos.services.invoice_service import InvoiceService

LOG = logging.getLogger(__name__)


@click.group()
def invoice():
    """Managing invoices"""


@invoice.command()
def generate():
    """Generate invoices"""
    click.secho("Generating invoice...", fg="blue")
    service = InvoiceService()
    invoice_url = service.create_invoice()
    click.secho("Invoice generated", fg="green")
    click.secho(f"URL: {invoice_url}", fg="blue")



from typing import Iterable

from click.shell_completion import CompletionItem

from edos.cache import cache
from edos.services.database_service import DatabaseService


def create_completion_from_names(names: Iterable[str], incomplete: str):
    return [CompletionItem(name) for name in names if name.lower().startswith(incomplete.lower())]


def cluster_id_completion(ctx, args, incomplete):
    @cache.memoize(expire=10)
    def get_names():
        return [*DatabaseService().get_clusters()]

    return create_completion_from_names(get_names(), incomplete)


def get_new_secret_name(old_name: str):
    """
    secret names are typically in this format:
        <secret>-<version> where version is a number
    this function will get a new secret with incremented version
    if version is not present, append it
    :param old_name: <secret>-<version>
    :return: <secret>-<version+1>
    """
    split_secret = old_name.rsplit("-", 1)
    if "-" not in old_name or not split_secret[1].isnumeric():
        return old_name + "-1"
    new_secret_version = int(split_secret[1]) + 1
    return f"{split_secret[0]}-{new_secret_version}"



import logging
import os
import uuid
from datetime import datetime

import click

from edos.cli.common import cluster_id_completion
from edos.exceptions import UserReadableException
from edos.services.database_service import DatabaseService
from edos.settings import conf

LOG = logging.getLogger(__name__)


@click.group()
def database():
    """Managing Digital Ocean databases"""


@database.command()
@click.argument("cluster_id", shell_complete=cluster_id_completion)
def ls(cluster_id):
    """command to get databases in certain cluster"""
    service = DatabaseService()
    databases = service.get_databases(cluster_id)
    for db in databases:
        click.echo(db)


@database.command()
@click.argument("cluster_id", shell_complete=cluster_id_completion)
@click.argument("name")
@click.option("--existing-user", type=str, required=False, help="Use existing user instead of creating a new one")
@click.option("--from-dump-file", type=click.Path(exists=True), required=False, help="SQL file to import")
@click.option("--no-confirm", is_flag=True, help="Do not ask for confirmation")
@click.option("--upsert", is_flag=True, help="If database already exists, use it instead of failing")
def create(cluster_id, name, existing_user = None, from_dump_file = None, no_confirm = False, upsert = False):
    """command for create a database in certain cluster"""
    service = DatabaseService()
    if not no_confirm and not click.confirm(f"Are you sure, that you want to create {name} " f"database in {cluster_id} cluster?"):
        return click.echo("creation cancelled")

    db_url = service.create_database(cluster_id, name, existing_user=existing_user, from_dump_file=from_dump_file, upsert=upsert)
    click.echo(f'DB url =>     "{db_url}"')


@database.command()
@click.argument("cluster_id", shell_complete=cluster_id_completion, type=str)
@click.argument("db_name", type=str)
@click.argument("db_user", type=str, required=False)
def dump(cluster_id, db_name, db_user):
    """
    command for create db snapshot and save it into yours /tmp,
    db_user is optional (default is db_name)
    """
    service = DatabaseService()
    cluster = service.get_cluster(cluster_id)
    _user = db_user if db_user else db_name
    user = service.get_database_user(cluster.id, _user)
    server_dumpfile_name = f"dump_{uuid.uuid4()}.sql"
    if cluster.engine == "pg":
        dump_command = (
            f"ssh {conf.SSH_CONNECTION} 'PGPASSWORD={user.password} pg_dump -U "
            f"{user.name} -h {cluster.privateConnection.host} -p "
            f"{cluster.privateConnection.port} -d {db_name} > "
            f"/tmp/{server_dumpfile_name}'"
        )
    elif cluster.engine == "mysql":
        dump_command = (
            f"ssh {conf.SSH_CONNECTION} 'mysqldump -u {user.name} -p{user.password} "
            f"-h {cluster.privateConnection.host} -P {cluster.privateConnection.port} "
            f"{db_name} > /tmp/{server_dumpfile_name}'"
        )
    else:
        raise UserReadableException("Bad cluster. Only PG and Mysql clusters are supported")
    timestamp: str = str(datetime.now().timestamp())
    local_filename = f"{db_name}_{timestamp}.sql"
    copy_command = f"scp {conf.SSH_CONNECTION}:/tmp/{server_dumpfile_name} /tmp/{local_filename}"
    remove_command = f"ssh {conf.SSH_CONNECTION} rm /tmp/{server_dumpfile_name}"
    click.echo("Dumping database")
    os.system(dump_command)
    os.system(copy_command)
    os.system(remove_command)
    click.echo(f"Database was successfully dumped. You can find it in " f"your /tmp/{local_filename}")


@database.command()
@click.argument("cluster_id", shell_complete=cluster_id_completion)
@click.argument("database_name", type=str)
@click.argument("username", type=str)
@click.option("--input-file", type=click.Path(exists=True), required=False, help="SQL file to execute")
def psql(cluster_id, database_name, username, input_file=None):
    """command for running psql shell"""
    service = DatabaseService()
    cluster = service.get_cluster(cluster_id)
    if cluster.engine == "pg":
        service.run_psql_shell(cluster, database_name, username, input_file)
    else:
        raise UserReadableException("Bad cluster. Only PG clusters are supported")

# @database.command()
# @click.argument('cluster_id')
# @click.argument('name')
# def recreate(cluster_id, name):
#     """command for recreate database
#     (when it needs to be dropped and created again)"""
#     service = DatabaseService()
#     if not confirm_action(f"Are you sure, that you want to
#     recreate {name} database in {cluster_id} cluster?"):
#         return click.echo('recreatin cancelled')
#     service.recreate_database(cluster_id, name)
#     return click.echo('successfully recreated database')



import configparser
import logging
import os

import yaml
from appdirs import AppDirs

from ..exceptions import UserReadableException


class LocalConfig(dict):
    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            raise UserReadableException(
                "Error: DO setup tool is not configured. Run {} first.".format("edos-configure")
            )


def _read_config(config_path):
    conf = configparser.ConfigParser()
    conf.read(config_path)
    conf = LocalConfig((s, dict(conf.items(s))) for s in conf.sections())
    return conf


class Config:
    LOG = logging.getLogger(__name__)

    def __init__(self):
        self.reload()

    def reload(self):
        if os.environ.get("TEST_MODE", False) == "true":
            self.config = {
                "DO": {"token": "test_token"},
                "AWS": {"aws_access_key": "test_key", "aws_secret_key": "test_secret"},
                "SWARMPIT": {"token": "test_token"},
                "INVOICE": {
                    "fakturoid_api_key": "test_key",
                    "clockify_api_key": "test_key",
                    "fakturoid_account_email": "test_account",
                    "fakturoid_account_slug": "test_slug",
                },
                                "SSH": {"user": "administrator", "hostname": "swarm1"},
            }
        else:
            self.config = _read_config(self.CONFIG_PATH) if os.path.exists(self.CONFIG_PATH) else LocalConfig()

        self.DIGITAL_OCEAN_CONFIG = self.config["DO"] if "DO" in self.config else {}
        self.AWS_CONFIG = self.config["AWS"] if "AWS" in self.config else {}
        self.SWARMPIT_CONFIG = self.config["SWARMPIT"] if "SWARMPIT" in self.config else {}
        self.INVOICE_CONFIG = self.config["INVOICE"] if "INVOICE" in self.config else {}
        self.SSH_CONFIG = self.config["SSH"] if "SSH" in self.config else {}

    # Interactive mode
    INTERACTIVE = bool(yaml.full_load(os.getenv("INTERACTIVE", "True")))

    USER_DIR = AppDirs("Endevel", "edos")
    CONFIG_PATH = os.path.join(USER_DIR.user_config_dir, "config.ini")

    AWS_ENDPOINT_URL = "https://fra1.digitaloceanspaces.com"
    AWS_REGION = "fra1"

    DOCKER_MAIN_SWARM_HOSTNAME = "swarm1"

    CLOCKIFY_ENDEVEL_WORKSPACE_ID = "5d19eacd1080ec307ed7ae12"
    CLOCKIFY_API_URL = "api.clockify.me/v1"
    FAKTUROID_USER_AGENT = "EDOS (stepan.binko@endevel.cz)"
    ENDEVEL_ICO = "11769327"

    PSQL_CLUSTER_ID = "c26441b7-d10a-453e-a367-ea2cdc27ea07"
    PSQL_DEV_CLUSTER_ID = "92ae8f86-bf35-4129-979a-de0889f7f5d4"
    MYSQL_CLUSTER_ID = "5ab58201-4008-4faf-bc38-f983633adef9"

    @property
    def SSH_USER(self):
        return self.SSH_CONFIG.get("user", "administrator")

    @property
    def SSH_HOSTNAME(self):
        return self.SSH_CONFIG.get("hostname", "swarm1")

    @property
    def SSH_CONNECTION(self):
        return f"{self.SSH_USER}@{self.SSH_HOSTNAME}"

    @property
    def DOCKER_SSH_BASE_URL(self):
        return f"ssh://{self.SSH_CONNECTION}"



import logging

import click
import yaml

from edos.cli.common import cluster_id_completion
from edos.services.database_service import DatabaseService

LOG = logging.getLogger(__name__)


@click.group()
def cluster():
    """Database cluster management"""


@cluster.command()
def ls():
    """List clusters - IDs and names"""
    cluster_dict = DatabaseService().get_clusters()
    for key, value in cluster_dict.items():
        click.echo(f"{key} - {value}")


@cluster.command()
@click.argument("cluster_id", shell_complete=cluster_id_completion)
def inspect(cluster_id: str):
    """Inspect cluster"""
    cl = DatabaseService().get_cluster(cluster_id)
    cl_dict = {
        "id": cl.id,
        "name": cl.name,
        "status": cl.status,
        "engine": cl.engine,
        "version": cl.version,
        "size": cl.size,
        "tags": cl.tags,
        "databases": [db.name for db in cl.db_names] if cl.db_names is not None else "N/A",
        "users": [user.name for user in cl.users] if cl.users is not None else "N/A",
    }
    click.echo_via_pager(yaml.dump(cl_dict, indent=4, sort_keys=False, default_flow_style=False))



import dataclasses

from edos.api.swarmpit.models.memory_response import MemoryResponse
from edos.api.swarmpit.services import ServiceSwarmpitAPI


@dataclasses.dataclass
class MemoryServiceStats:
    service_name: str
    memory_reservation: int
    actual_memory_usage: int


class SwarmpitService:
    def __init__(self):
        self.api = ServiceSwarmpitAPI()

    def get_service_memory_stats(self) -> list[MemoryServiceStats]:
        timeseries: list[MemoryResponse] = self.api.get_memory_timeseries()

        service_stats = []
        for item in timeseries:
            memory_usage = int(item.actual_memory_usage / 1000000)  # MiB
            service_stats.append(
                MemoryServiceStats(
                    service_name=item.service_name,
                    memory_reservation=item.memory_reservation,
                    actual_memory_usage=memory_usage,
                )
            )
        return service_stats



from edos.api.digitalocean.databases.models.cluster import DatabaseCluster
from edos.api.digitalocean.databases.models.connections import DatabaseConnection
from edos.api.digitalocean.databases.models.database import Database
from edos.api.digitalocean.databases.models.database_user import DatabaseUser


def parse_one_cluster(cluster: dict) -> DatabaseCluster:
    return DatabaseCluster(
        id=cluster.get("id"),
        name=cluster.get("name"),
        engine=cluster.get("engine"),
        version=cluster.get("version"),
        num_nodes=int(cluster.get("num_nodes", 0)),
        size=cluster.get("size"),
        region=cluster.get("region"),
        status=cluster.get("status"),
        created_at=cluster.get("created_at"),
        private_network_uuid=cluster.get("private_network_uuid"),
        tags=cluster.get("tags", None),
        db_names=[Database(db) for db in (cluster.get("db_names", []) if cluster.get("db_names", []) else [])],
        connection=parse_one_connection(cluster.get("connection")),
        privateConnection=parse_one_connection(cluster.get("private_connection")),
        users=[parse_one_user(user) for user in (cluster.get("users", []) if cluster.get("users", []) else [])],
    )


def parse_one_database(database: dict) -> Database:
    return Database(name=database.get("name"))


def parse_one_user(user: dict) -> DatabaseUser:
    return DatabaseUser(
        name=user.get("name"),
        role=user.get("role"),
        password=user.get("password"),
    )


def parse_one_connection(connection: dict) -> DatabaseConnection:
    return DatabaseConnection(
        uri=connection.get("uri"),
        database=connection.get("database"),
        host=connection.get("host"),
        port=int(connection.get("port")),
        user=connection.get("user"),
        password=connection.get("password"),
        ssl=connection.get("ssl"),
    )
